#
# core.py
#
# Copyright (C) 2009 None <terror.macbeth.I@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# Deluge is free software.
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# deluge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with deluge.    If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.
#
#    In addition, as a special exception, the copyright holders give
#    permission to link the code of portions of this program with the OpenSSL
#    library.
#    You must obey the GNU General Public License in all respects for all of
#    the code used other than OpenSSL. If you modify file(s) with this
#    exception, you may extend this exception to your version of the file(s),
#    but you are not obligated to do so. If you do not wish to do so, delete
#    this exception statement from your version. If you delete this exception
#    statement from all source files in the program, then also delete it here.
#

from deluge.log import LOG as log
from deluge.plugins.pluginbase import CorePluginBase
import deluge.component as component
import deluge.configmanager
from deluge.core.rpcserver import export
from twisted.internet.task import LoopingCall


DEFAULT_PREFS = {
    "highest_count": 2,         # number of files that get highest priority
    "high_count": 3,            # number of files that get high priority
    "label": "autopriority",    # torrents with that label are processed
    "update_interval": 60,
    "priorities_mapping": {"highest": 7,
                           "high": 5,
                           "norm": 1,
                           "dnd": 0},
}

class Core(CorePluginBase):
    def enable(self):
        self.config = deluge.configmanager.ConfigManager("autopriority.conf",
                                                         DEFAULT_PREFS)
        self.core = component.get("Core")
        # self.files_state = {torrent_id: [(progress[0], priorities[0]), ... ]}
        self.files_state = {}

        self.highest = self.config["highest_count"]
        self.high = self.config["high_count"]
        self.interval = self.config["update_interval"]
        self.label = self.config["label"]
        self.priorities_mapping = self.config["priorities_mapping"]

        self.update_status_timer = LoopingCall(self.get_progress)
        self.update_status_timer.start(self.interval)

    def disable(self):
        self.update_status_timer.stop()

    def update(self):
        pass

    def get_progress(self):
        """
        Keeps track of newly finished files and external priority changes.
        If either of this two occur, self.update_priorities is called for
        the corresponding torrent.
        """
        torrents = self.core.get_torrents_status({"state": "Downloading"},
                                                 ["file_progress",
                                                  "files",
                                                  "file_priorities",
                                                  "label"])
        for torrent, value_dict in torrents.items():
            if value_dict["label"] != self.label:
                torrents.pop(torrent)
        log.debug("AutoPriority: watching label %s" % self.label)
        log.debug("AutoPriority: watching %i torrent(s)" % len(torrents))
        for torrent in self.files_state.keys():
            # remove torrents no longer "Downloading" from cache
            if not torrents.has_key(torrent):
                self.files_state.pop(torrent)
        for torrent, values in torrents.items():
            now_finished = [(x == 1.0) for x in values["file_progress"]]
            current_state = zip(now_finished, values["file_priorities"])
            if not self.files_state.has_key(torrent):
                # torrent seen for the first time:
                self.files_state[torrent] = current_state
                self.update_priorities(torrent, values)
            elif self.files_state[torrent] != current_state:
                # a file finished downloading or ui changed priorities:
                self.files_state[torrent] = current_state
                self.update_priorities(torrent, values)

    def update_priorities(self, torrent, values):
        """
        Updates the file priorities for the given torrent.
        Files are sorted alphabeticaly.
        If a file is neither finished nor set to "Do Not Download" the
        status is changed to "Highest" or "High" if applicable.
        """
        unsorted_names = [f["path"].lower() for f in values["files"]]
        sorted_names = unsorted_names[:]
        sorted_names.sort()
        highest = self.highest
        high = self.high
        P = self.priorities_mapping
        priorities = list(values["file_priorities"])
        for filename in sorted_names:
            i = unsorted_names.index(filename)
            progress = values["file_progress"]
            if progress[i] < 1.0 and priorities[i] != P["dnd"]:
                if highest:
                    priorities[i] = P["highest"]
                    highest -= 1
                elif high:
                    priorities[i] = P["high"]
                    high -= 1
                else:
                    break
        self.core.set_torrent_file_priorities(torrent, priorities)

    @export
    def set_config(self, config):
        "sets the config dictionary"
        for key in config.keys():
            self.config[key] = config[key]
        # update members to new config
        self.highest = self.config["highest_count"]
        self.high = self.config["high_count"]
        self.interval = self.config["update_interval"]
        self.label = self.config["label"]

        # clear cache and reregister timer callback
        self.files_state = {}
        self.update_status_timer.stop()
        self.update_status_timer.start(self.interval)

        self.config.save()

    @export
    def get_config(self):
        "returns the config dictionary"
        return self.config.config
